@@ -28,21 +28,31 @@ import java.io.InputStreamReader;
 import java.io.PrintWriter;
 import java.lang.management.ManagementFactory;
 import java.lang.reflect.InvocationTargetException;
-import java.net.URL;
+import java.nio.charset.StandardCharsets;
+import java.nio.file.Path;
+import java.nio.file.Paths;
+import java.security.AccessControlContext;
+import java.security.AccessController;
 import java.security.Policy;
-import java.util.Enumeration;
+import java.security.PrivilegedAction;
+import java.security.ProtectionDomain;
+import java.security.Provider;
+import java.security.Security;
+import java.util.ArrayList;
+import java.util.Collections;
 import java.util.Iterator;
+import java.util.List;
+import java.util.Properties;
+import java.util.ServiceConfigurationError;
 import java.util.ServiceLoader;
-import java.util.jar.Attributes;
-import java.util.jar.JarFile;
 import java.util.logging.LogManager;
 
-import java.util.jar.Manifest;
 import java.util.prefs.Preferences;
 import java.util.regex.Matcher;
 import java.util.regex.Pattern;
 
 import org.jboss.modules.log.JDKModuleLogger;
+import org.jboss.modules.log.StreamModuleLogger;
 
 import static java.security.AccessController.doPrivileged;
 import static org.jboss.modules.SecurityActions.setContextClassLoader;
@@ -87,6 +97,7 @@ public final class Main {
         System.out.println("                  A list of module dependencies to add to the class path;");
         System.out.println("                  requires -class or -cp");
         System.out.println("    -deptree      Print the dependency tree of the given module instead of running it");
+        System.out.println("    -debuglog     Enable debug mode output to System.out during bootstrap before any logging manager is installed");
         System.out.println("    -jar          Specify that the final argument is the name of a");
         System.out.println("                  JAR file to run as a module; not compatible with -class");
         System.out.println("    -jaxpmodule <module-spec>");
@@ -94,6 +105,8 @@ public final class Main {
         System.out.println("    -secmgr       Run with a security manager installed; not compatible with -secmgrmodule");
         System.out.println("    -secmgrmodule <module-spec>");
         System.out.println("                  Run with a security manager module; not compatible with -secmgr");
+        System.out.println("    -add-provider <module-name>[/<class-name>]");
+        System.out.println("                  Add a security provider of the given module and class (can be given more than once)");
         System.out.println("    -addindex     Specify that the final argument is a");
         System.out.println("                  jar to create an index for");
         System.out.println("    -modify       Modify the indexes jar in-place");
@@ -118,12 +131,14 @@ public final class Main {
         boolean classDefined = false;
         boolean depTree = false;
         String nameArgument = null;
-        ModuleIdentifier jaxpModuleIdentifier = null;
+        String jaxpModuleName = null;
         boolean defaultSecMgr = false;
         String secMgrModule = null;
         boolean addIndex = false;
         boolean modifyInPlace = false;
-        for (int i = 0, argsLength = argsLen; i < argsLength; i++) {
+        boolean debuglog = false;
+        final List<String> addedProviders = new ArrayList<>();
+        for (int i = 0; i < argsLen; i++) {
             final String arg = args[i];
             try {
                 if (arg.charAt(0) == '-') {
@@ -166,8 +181,10 @@ public final class Main {
                             System.exit(1);
                         }
                         depTree = true;
+                    } else if ("-debuglog".equals(arg)) {
+                        debuglog = true;
                     } else if ("-jaxpmodule".equals(arg)) {
-                        jaxpModuleIdentifier = ModuleIdentifier.fromString(args[++i]);
+                        jaxpModuleName = args[++i];
                     } else if ("-jar".equals(arg)) {
                         if (jar) {
                             System.err.println("-jar flag may only be specified once");
@@ -252,7 +269,9 @@ public final class Main {
                             System.err.println("-secmgrmodule may not be specified when -secmgr is given");
                             System.exit(1);
                         }
-                        secMgrModule = args[++i];
+                        secMgrModule = args[++ i];
+                    } else if ("-add-provider".equals(arg)) {
+                        addedProviders.add(args[++ i]);
                     } else {
                         System.err.printf("Invalid option '%s'\n", arg);
                         usage();
@@ -272,7 +291,7 @@ public final class Main {
                 System.exit(1);
             }
         }
-        
+
         if (modifyInPlace && ! addIndex) {
             System.err.println("-modify requires -addindex");
             usage();
@@ -280,6 +299,7 @@ public final class Main {
         }
 
         if (addIndex) {
+            System.err.println("WARNING: -addindex is deprecated and may be removed in a future release");
             if (nameArgument == null) {
                 System.err.println("-addindex requires a target JAR name");
                 usage();
@@ -290,8 +310,8 @@ public final class Main {
                 usage();
                 System.exit(1);
             }
-            if (jaxpModuleIdentifier != null) {
-                System.err.println("-jaxpModuleIdentifier may not be used with -addindex");
+            if (jaxpModuleName != null) {
+                System.err.println("-jaxpModuleName may not be used with -addindex");
                 usage();
                 System.exit(1);
             }
@@ -325,6 +345,9 @@ public final class Main {
                 usage();
                 System.exit(1);
             }
+            if (! addedProviders.isEmpty()) {
+                System.err.println("-add-provider may not be used with -addindex");
+            }
             if (depTree) {
                 System.err.println("-deptree may not be used with -addindex");
                 usage();
@@ -354,34 +377,64 @@ public final class Main {
         }
 
         if (depTree) {
-            DependencyTreeViewer.print(new PrintWriter(System.out), ModuleIdentifier.fromString(nameArgument), LocalModuleFinder.getRepoRoots(true));
+            DependencyTreeViewer.print(new PrintWriter(System.out), nameArgument, LocalModuleFinder.getRepoRoots(true));
             System.exit(0);
         }
 
+        if(debuglog) {
+            // Install the StreamModuleLogger on System.out to capture bootstrap messages
+            Module.setModuleLogger(new StreamModuleLogger(System.out));
+        }
+
         final ModuleLoader loader;
         final ModuleLoader environmentLoader;
         environmentLoader = DefaultBootModuleLoaderHolder.INSTANCE;
-        final ModuleIdentifier moduleIdentifier;
+        final Path rootPath = Paths.get("").toAbsolutePath();
+        final String moduleName;
+        final String className;
         if (jar) {
-            loader = new JarModuleLoader(environmentLoader, new JarFile(nameArgument));
-            moduleIdentifier = ((JarModuleLoader) loader).getMyIdentifier();
+            loader = new ModuleLoader(new FileSystemClassPathModuleFinder(environmentLoader));
+            moduleName = rootPath.resolve(nameArgument).normalize().toString();
+            className = null;
         } else if (classpathDefined || classDefined) {
-            loader = new ClassPathModuleLoader(environmentLoader, nameArgument, classpath, deps);
-            moduleIdentifier = ModuleIdentifier.CLASSPATH;
+            final String[] items;
+            if (classpath != null) {
+                AccessController.doPrivileged(new PropertyWriteAction("java.class.path", classpath));
+                items = classpath.split(Pattern.quote(File.pathSeparator));
+            } else {
+                items = NO_STRINGS;
+            }
+            for (int i = 0; i < items.length; i++) {
+                items[i] = rootPath.resolve(items[i]).normalize().toString();
+            }
+            loader = new DelegatingModuleLoader(environmentLoader, new ClassPathModuleFinder(environmentLoader, items, deps, nameArgument));
+            moduleName = "<classpath>";
+            className = null;
         } else {
             loader = environmentLoader;
-            moduleIdentifier = ModuleIdentifier.fromString(nameArgument);
+            final int idx = nameArgument.lastIndexOf('/');
+            if (idx != -1) {
+                moduleName = nameArgument.substring(0, idx);
+                className = nameArgument.substring(idx + 1);
+            } else {
+                moduleName = nameArgument;
+                className = null;
+            }
         }
-        Module.initBootModuleLoader(loader);
-        if (jaxpModuleIdentifier != null) {
-            __JAXPRedirected.changeAll(jaxpModuleIdentifier, Module.getBootModuleLoader());
+        Module.initBootModuleLoader(environmentLoader);
+        if (jaxpModuleName != null) {
+            __JAXPRedirected.changeAll(jaxpModuleName, Module.getBootModuleLoader());
         } else {
-            __JAXPRedirected.changeAll(moduleIdentifier, Module.getBootModuleLoader());
+            __JAXPRedirected.changeAll(moduleName, Module.getBootModuleLoader());
+        }
+
+        if (moduleName == null) {
+            return;
         }
 
         final Module module;
         try {
-            module = loader.loadModule(moduleIdentifier);
+            module = loader.loadModule(moduleName);
         } catch (ModuleNotFoundException e) {
             e.printStackTrace(System.err);
             System.exit(1);
@@ -389,10 +442,10 @@ public final class Main {
         }
         final String ourJavaVersion = doPrivileged(new PropertyReadAction("java.specification.version", "1.6"));
         final String requireJavaVersion = module.getProperty("jboss.require-java-version", ourJavaVersion);
-        final Pattern versionPattern = Pattern.compile("1\\.(\\d+)");
+        final Pattern versionPattern = Pattern.compile("(?:1\\.)?(\\d+)");
         final Matcher requireMatcher = versionPattern.matcher(requireJavaVersion);
         final Matcher ourMatcher = versionPattern.matcher(ourJavaVersion);
-        if (requireMatcher.matches() && ourMatcher.matches() && Integer.valueOf(requireMatcher.group(1)) > Integer.valueOf(ourMatcher.group(1))) {
+        if (requireMatcher.matches() && ourMatcher.matches() && Integer.parseInt(requireMatcher.group(1)) > Integer.parseInt(ourMatcher.group(1))) {
             System.err.printf("This application requires Java specification version %s or later to run (this Java virtual machine implements specification version %s)%n", requireJavaVersion, ourJavaVersion);
             System.exit(1);
         }
@@ -423,7 +476,7 @@ public final class Main {
         if (secMgrModule != null) {
             final Module loadedModule;
             try {
-                loadedModule = loader.loadModule(ModuleIdentifier.fromString(secMgrModule));
+                loadedModule = environmentLoader.loadModule(secMgrModule);
             } catch (ModuleNotFoundException e) {
                 e.printStackTrace(System.err);
                 System.exit(1);
@@ -481,22 +534,76 @@ public final class Main {
             ManagementFactory.getPlatformMBeanServer();
         }
 
+        for (String addedProvider : addedProviders) {
+            final int idx = addedProvider.indexOf('/');
+            if (idx != -1) {
+                final String provModule = addedProvider.substring(0, idx);
+                final String provClazz = addedProvider.substring(idx + 1);
+                final Class<? extends Provider> providerClass;
+                try {
+                    providerClass = Class.forName(provClazz, false, environmentLoader.loadModule(provModule).getClassLoaderPrivate()).asSubclass(Provider.class);
+                    // each provider needs permission to install itself
+                    doPrivileged(new AddProviderAction(providerClass.getConstructor().newInstance()), getProviderContext(providerClass));
+                } catch (Exception e) {
+                    Module.getModuleLogger().trace(e, "Failed to initialize a security provider");
+                }
+            } else {
+                final ModuleClassLoader classLoader = environmentLoader.loadModule(addedProvider).getClassLoaderPrivate();
+                final ServiceLoader<Provider> providerServiceLoader = ServiceLoader.load(Provider.class, classLoader);
+                final Iterator<Provider> iterator = providerServiceLoader.iterator();
+                for (;;) try {
+                    if (! (iterator.hasNext())) {
+                        Module.getModuleLogger().trace("Module \"%s\" did not contain a security provider service", addedProvider);
+                        break;
+                    }
+                    final Provider provider = iterator.next();
+                    final Class<? extends Provider> providerClass = provider.getClass();
+                    // each provider needs permission to install itself
+                    doPrivileged(new AddProviderAction(provider), getProviderContext(providerClass));
+                } catch (ServiceConfigurationError | RuntimeException e) {
+                    Module.getModuleLogger().trace(e, "Failed to initialize a security provider");
+                }
+            }
+        }
+
+        final ServiceLoader<Provider> providerServiceLoader = ServiceLoader.load(Provider.class, bootClassLoader);
+        Iterator<Provider> iterator = providerServiceLoader.iterator();
+        for (;;) try {
+            if (! (iterator.hasNext())) break;
+            final Provider provider = iterator.next();
+            final Class<? extends Provider> providerClass = provider.getClass();
+            // each provider needs permission to install itself
+            doPrivileged(new AddProviderAction(provider), getProviderContext(providerClass));
+        } catch (ServiceConfigurationError | RuntimeException e) {
+            Module.getModuleLogger().trace(e, "Failed to initialize a security provider");
+        }
+
         ModuleLoader.installMBeanServer();
 
+        final ArrayList<String> argsList = new ArrayList<>(moduleArgs.length);
+        Collections.addAll(argsList, moduleArgs);
+
+        final ServiceLoader<PreMain> preMainServiceLoader = ServiceLoader.load(PreMain.class, bootClassLoader);
+        for (PreMain preMain : preMainServiceLoader) {
+            preMain.run(argsList);
+        }
+
         try {
-            module.run(moduleArgs);
+            final String[] argsArray = argsList.toArray(NO_STRINGS);
+            if (className != null) module.run(className, argsArray); else module.run(argsArray);
         } catch (InvocationTargetException e) {
             throw e.getCause();
         }
         return;
     }
 
+    private static AccessControlContext getProviderContext(final Class<? extends Provider> providerClass) {
+        return new AccessControlContext(new ProtectionDomain[] { providerClass.getProtectionDomain() });
+    }
+
     private static String getServiceName(ClassLoader classLoader, String className) throws IOException {
-        final InputStream stream = classLoader.getResourceAsStream("META-INF/services/" + className);
-        if (stream == null) {
-            return null;
-        }
-        try {
+        try (final InputStream stream = classLoader.getResourceAsStream("META-INF/services/" + className)) {
+            if ( stream == null ) return null;
             final BufferedReader reader = new BufferedReader(new InputStreamReader(stream));
             String line;
             while ((line = reader.readLine()) != null) {
@@ -509,8 +616,6 @@ public final class Main {
                 return line;
             }
             return null;
-        } finally {
-            StreamUtil.safeClose(stream);
         }
     }
 
@@ -518,29 +623,17 @@ public final class Main {
     private static final String VERSION_STRING;
 
     static {
-        final Enumeration<URL> resources;
+        Properties versionProps = new Properties();
         String jarName = "(unknown)";
         String versionString = "(unknown)";
-        try {
-            final ClassLoader classLoader = Main.class.getClassLoader();
-            resources = classLoader == null ? ModuleClassLoader.getSystemResources("META-INF/MANIFEST.MF") : classLoader.getResources("META-INF/MANIFEST.MF");
-            while (resources.hasMoreElements()) {
-                final URL url = resources.nextElement();
-                try {
-                    final InputStream stream = url.openStream();
-                    if (stream != null) try {
-                        final Manifest manifest = new Manifest(stream);
-                        final Attributes mainAttributes = manifest.getMainAttributes();
-                        if (mainAttributes != null && "JBoss Modules".equals(mainAttributes.getValue("Specification-Title"))) {
-                            jarName = mainAttributes.getValue("Jar-Name");
-                            versionString = mainAttributes.getValue("Jar-Version");
-                        }
-                    } finally {
-                        StreamUtil.safeClose(stream);
-                    }
-                } catch (IOException ignored) {}
+        try (InputStream stream = Main.class.getResourceAsStream("version.properties")) {
+            if (stream != null) try (InputStreamReader reader = new InputStreamReader(stream, StandardCharsets.UTF_8)) {
+                versionProps.load(reader);
+                jarName = versionProps.getProperty("jarName", jarName);
+                versionString = versionProps.getProperty("version", versionString);
             }
-        } catch (IOException ignored) {}
+        } catch (IOException ignored) {
+        }
         JAR_NAME = jarName;
         VERSION_STRING = versionString;
     }
@@ -562,4 +655,17 @@ public final class Main {
     public static String getVersionString() {
         return VERSION_STRING;
     }
+
+    static final class AddProviderAction implements PrivilegedAction<Void> {
+        private final Provider provider;
+
+        AddProviderAction(final Provider provider) {
+            this.provider = provider;
+        }
+
+        public Void run() {
+            Security.addProvider(provider);
+            return null;
+        }
+    }
 }
