@@ -22,6 +22,8 @@ import static java.security.AccessController.doPrivileged;
 import static org.jboss.modules.management.ObjectProperties.property;
 
 import java.lang.management.ManagementFactory;
+import java.lang.reflect.Method;
+import java.lang.reflect.Modifier;
 import java.lang.reflect.UndeclaredThrowableException;
 import java.security.PrivilegedAction;
 import java.security.PrivilegedActionException;
@@ -30,6 +32,7 @@ import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
 import java.util.Collections;
+import java.util.Comparator;
 import java.util.Iterator;
 import java.util.LinkedHashMap;
 import java.util.List;
@@ -38,6 +41,7 @@ import java.util.NoSuchElementException;
 import java.util.Set;
 import java.util.SortedMap;
 import java.util.TreeMap;
+import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicLong;
@@ -62,6 +66,7 @@ import javax.management.ObjectName;
  *
  * @author <a href="mailto:david.lloyd@redhat.com">David M. Lloyd</a>
  * @author <a href="mailto:jbailey@redhat.com">John Bailey</a>
+ * @author <a href="mailto:ropalka@redhat.com">Richard Opalka</a>
  * @author Jason T. Greene
  *
  * @apiviz.landmark
@@ -75,6 +80,7 @@ public class ModuleLoader {
     private static final RuntimePermission MODULE_ITERATE_PERM = new RuntimePermission("canIterateModules");
 
     private static final AtomicInteger SEQ = new AtomicInteger(1);
+    private static final Class<?>[] JUST_MODULE_IDENTIFIER = { ModuleIdentifier.class };
 
     private static volatile MBeanReg REG_REF = new TempMBeanReg();
 
@@ -83,7 +89,7 @@ public class ModuleLoader {
      */
     public static final ModuleFinder[] NO_FINDERS = new ModuleFinder[0];
 
-    private final ConcurrentMap<ModuleIdentifier, FutureModule> moduleMap = new UnlockedReadHashMap<ModuleIdentifier, FutureModule>(256);
+    private final ConcurrentMap<String, FutureModule> moduleMap = new ConcurrentHashMap<>();
     private final ModuleFinder[] finders;
 
     private final boolean canRedefine;
@@ -96,6 +102,9 @@ public class ModuleLoader {
     private final AtomicInteger raceCount = new AtomicInteger();
     private final AtomicInteger classCount = new AtomicInteger();
 
+    private final boolean overridesFindModuleByIdentifier;
+    private final boolean overridesPreloadModuleByIdentifier;
+
     ModuleLoader(boolean canRedefine, boolean skipRegister) {
         this(canRedefine, skipRegister, NO_FINDERS);
     }
@@ -121,6 +130,32 @@ public class ModuleLoader {
                 return null;
             }
         });
+        // determine if the subclass overrides the crusty old find-module-by-identifier method
+        boolean foundFindModule = false;
+        boolean foundPreloadModule = false;
+        for (Class<? extends ModuleLoader> clazz = getClass(); clazz != ModuleLoader.class; clazz = clazz.getSuperclass().asSubclass(ModuleLoader.class)) {
+            for (Method method : clazz.getDeclaredMethods()) {
+                if ((method.getModifiers() & (Modifier.STATIC | Modifier.PRIVATE)) == 0) {
+                    if (method.getName().equals("findModule")) {
+                        if (ModuleSpec.class.isAssignableFrom(method.getReturnType())) {
+                            if (Arrays.equals(method.getParameterTypes(), JUST_MODULE_IDENTIFIER)) {
+                                foundFindModule = true;
+                                if (foundPreloadModule) break;
+                            }
+                        }
+                    } else if (method.getName().equals("preloadModule")) {
+                        if (Module.class.isAssignableFrom(method.getReturnType())) {
+                            if (Arrays.equals(method.getParameterTypes(), JUST_MODULE_IDENTIFIER)) {
+                                foundPreloadModule = true;
+                                if (foundFindModule) break;
+                            }
+                        }
+                    }
+                }
+            }
+        }
+        overridesFindModuleByIdentifier = foundFindModule;
+        overridesPreloadModuleByIdentifier = foundPreloadModule;
     }
 
     /**
@@ -139,6 +174,15 @@ public class ModuleLoader {
         this(checkPermissions(), false, safeClone(finders));
     }
 
+    /**
+     * Construct a new instance.
+     *
+     * @param finder the single module finder to search (must not be {@code null})
+     */
+    public ModuleLoader(final ModuleFinder finder) {
+        this(new ModuleFinder[] { finder });
+    }
+
     private static ModuleFinder[] safeClone(ModuleFinder[] finders) {
         if (finders == null || finders.length == 0) {
             return NO_FINDERS;
@@ -205,22 +249,49 @@ public class ModuleLoader {
         return String.format("%s@%x for finders %s", getClass().getSimpleName(), hashCode(), Arrays.toString(finders));
     }
 
+    /**
+     * Get a string representation of the given module, used in debug output and stack traces.  This method may be
+     * overridden to provide a more detailed description.  By default it returns a string consisting of the module's
+     * name and version (if any).
+     *
+     * @param module the module to describe
+     * @return the description string
+     */
+    public String getModuleDescription(final Module module) {
+        final Version version = module.getVersion();
+        return version == null ? module.getName() : module.getName() + "@" + version;
+    }
+
     static void installMBeanServer() {
         REG_REF.installReal();
     }
 
     /**
-     * Load a module based on an identifier.  This method delegates to {@link #preloadModule(ModuleIdentifier)} and then
+     * Load a module based on an identifier.  This method delegates to {@link #preloadModule(String)} and then
      * links the returned module if necessary.
      *
      * @param identifier The module identifier
      * @return The loaded Module
      * @throws ModuleLoadException if the Module can not be loaded
+     * @deprecated Use {@link #loadModule(String)} instead.
      */
+    @Deprecated
     public final Module loadModule(ModuleIdentifier identifier) throws ModuleLoadException {
-        final Module module = preloadModule(identifier);
+        return loadModule(identifier.toString());
+    }
+
+    /**
+     * Load a module based on an identifier.  This method delegates to {@link #preloadModule(String)} and then
+     * links the returned module if necessary.
+     *
+     * @param name The module name
+     * @return The loaded Module
+     * @throws ModuleLoadException if the Module can not be loaded
+     */
+    public final Module loadModule(String name) throws ModuleLoadException {
+        final Module module = preloadModule(name);
         if (module == null) {
-            throw new ModuleNotFoundException(identifier.toString());
+            throw new ModuleNotFoundException(name);
         }
         module.relinkIfNecessary();
         return module;
@@ -233,15 +304,31 @@ public class ModuleLoader {
      * @param recursive {@code true} to find recursively nested modules, {@code false} to only find immediately nested modules
      * @return an iterator for the modules in this module finder
      * @throws SecurityException if the caller does not have permission to iterate module loaders
+     * @deprecated Use {@link #iterateModules(String, boolean)} instead.
      */
+    @Deprecated
     public final Iterator<ModuleIdentifier> iterateModules(final ModuleIdentifier baseIdentifier, final boolean recursive) {
+        return IteratorUtils.transformingIterator(iterateModules(baseIdentifier == null ? (String) null : baseIdentifier.toString(), recursive), ModuleIdentifier::fromString);
+    }
+
+    /**
+     * Iterate the modules which can be located via this module loader.
+     *
+     * @param baseName the identifier to start with, or {@code null} to iterate all modules; ignored if this module
+     * loader does not have a concept of nested modules
+     * @param recursive {@code true} to find recursively nested modules, {@code false} to only find immediately nested
+     * modules; ignored if this module finder does not have a concept of nested modules
+     * @return an iterator for the modules in this module finder
+     * @throws SecurityException if the caller does not have permission to iterate module loaders
+     */
+    public final Iterator<String> iterateModules(final String baseName, final boolean recursive) {
         final SecurityManager sm = System.getSecurityManager();
         if (sm != null) {
             sm.checkPermission(MODULE_ITERATE_PERM);
         }
-        return new Iterator<ModuleIdentifier>() {
+        return new Iterator<String>() {
             int idx;
-            Iterator<ModuleIdentifier> nested;
+            Iterator<String> nested;
 
             public boolean hasNext() {
                 for (;;) {
@@ -251,7 +338,7 @@ public class ModuleLoader {
                         }
                         final ModuleFinder finder = finders[idx++];
                         if (finder instanceof IterableModuleFinder) {
-                            nested = ((IterableModuleFinder) finder).iterateModules(baseIdentifier, recursive);
+                            nested = ((IterableModuleFinder) finder).iterateModules(baseName, recursive);
                         }
                     }
 
@@ -263,7 +350,7 @@ public class ModuleLoader {
                 }
             }
 
-            public ModuleIdentifier next() {
+            public String next() {
                 if (! hasNext()) {
                     throw new NoSuchElementException();
                 }
@@ -277,27 +364,77 @@ public class ModuleLoader {
     }
 
     /**
-     * Preload a module based on an identifier.  By default, no delegation is done and this method simply invokes
-     * {@link #loadModuleLocal(ModuleIdentifier)}.  A delegating module loader may delegate to the appropriate module
-     * loader based on loader-specific criteria (via the {@link #preloadModule(ModuleIdentifier, ModuleLoader)} method).
+     * This compatibility method just calls {@code preloadModule(methodIdentifier.toString())} and is usually not used;
+     * if however a legacy subclass overrides this method, then it will be called instead of {@link #preloadModule(String)},
+     * and that method will redirect to this one.
      *
-     * @param identifier the module identifier
+     * @deprecated Use {@link #preloadModule(String)} instead.
+     *
+     * @param identifier the module identifier (must not be {@code null})
      * @return the load result, or {@code null} if the module is not found
      * @throws ModuleLoadException if an error occurs
      */
+    @Deprecated
     protected Module preloadModule(ModuleIdentifier identifier) throws ModuleLoadException {
-        return loadModuleLocal(identifier);
+        if (overridesPreloadModuleByIdentifier) {
+            return loadModuleLocal(identifier);
+        } else {
+            return preloadModule(identifier.toString());
+        }
     }
 
     /**
-     * Preload an "exported" module based on an identifier.  By default this simply delegates to {@link #preloadModule(ModuleIdentifier)}.
+     * Preload a module based on an identifier.  By default, delegation is only done for system and this method otherwise simply invokes
+     * {@link #loadModuleLocal(String)}.  A delegating module loader may delegate to the appropriate module
+     * loader based on loader-specific criteria (via the {@link #preloadModule(String, ModuleLoader)} method).
+     *
+     * @param name the module identifier
+     * @return the load result, or {@code null} if the module is not found
+     * @throws ModuleLoadException if an error occurs
+     */
+    protected Module preloadModule(String name) throws ModuleLoadException {
+        if (overridesPreloadModuleByIdentifier) {
+            return preloadModule(ModuleIdentifier.fromString(name));
+        } else {
+            return loadModuleLocal(name);
+        }
+    }
+
+    /**
+     * Preload an "exported" module based on an identifier.  By default this simply delegates to {@link #preloadModule(String)}.
+     *
+     * @param name the module name
+     * @return the load result, or {@code null} if the module is not found
+     * @throws ModuleLoadException if an error occurs
+     */
+    protected Module preloadExportedModule(String name) throws ModuleLoadException {
+        return preloadModule(name);
+    }
+
+    /**
+     * Preload an "exported" module based on an identifier.  By default this simply delegates to {@link #preloadModule(String)}.
      *
      * @param identifier the module identifier
      * @return the load result, or {@code null} if the module is not found
      * @throws ModuleLoadException if an error occurs
+     * @deprecated Use {@link #preloadModule(String)} instead.
      */
+    @Deprecated
     protected Module preloadExportedModule(ModuleIdentifier identifier) throws ModuleLoadException {
-        return preloadModule(identifier);
+        return preloadModule(identifier.getName());
+    }
+
+    /**
+     * Utility method to delegate to another module loader, accessible from subclasses.  The delegate module loader
+     * will be queried for exported modules only.
+     *
+     * @param name the module name
+     * @param moduleLoader the module loader to delegate to
+     * @return the delegation result
+     * @throws ModuleLoadException if an error occurs
+     */
+    protected static Module preloadModule(String name, ModuleLoader moduleLoader) throws ModuleLoadException {
+        return moduleLoader.preloadExportedModule(name);
     }
 
     /**
@@ -308,9 +445,11 @@ public class ModuleLoader {
      * @param moduleLoader the module loader to delegate to
      * @return the delegation result
      * @throws ModuleLoadException if an error occurs
+     * @deprecated Use {@link #preloadModule(String, ModuleLoader)} instead.
      */
+    @Deprecated
     protected static Module preloadModule(ModuleIdentifier identifier, ModuleLoader moduleLoader) throws ModuleLoadException {
-        return moduleLoader.preloadExportedModule(identifier);
+        return moduleLoader.preloadExportedModule(identifier.toString());
     }
 
     /**
@@ -318,18 +457,33 @@ public class ModuleLoader {
      * module may not yet be resolved.  The returned module may have a different name than the given identifier if
      * the identifier is an alias for another module.
      *
-     * @param identifier the module identifier
+     * @param moduleIdentifier the module identifier
+     * @return the module
+     * @throws ModuleLoadException if an error occurs while loading the module
+     * @deprecated Use {@link #loadModuleLocal(String)} instead.
+     */
+    @Deprecated
+    protected final Module loadModuleLocal(ModuleIdentifier moduleIdentifier) throws ModuleLoadException {
+        return loadModuleLocal(moduleIdentifier.toString());
+    }
+
+    /**
+     * Try to load a module from this module loader.  Returns {@code null} if the module is not found.  The returned
+     * module may not yet be resolved.  The returned module may have a different name than the given identifier if
+     * the identifier is an alias for another module.
+     *
+     * @param name the module name
      * @return the module
      * @throws ModuleLoadException if an error occurs while loading the module
      */
-    protected final Module loadModuleLocal(ModuleIdentifier identifier) throws ModuleLoadException {
-        FutureModule futureModule = moduleMap.get(identifier);
+    protected final Module loadModuleLocal(String name) throws ModuleLoadException {
+        FutureModule futureModule = moduleMap.get(name);
         if (futureModule != null) {
             return futureModule.getModule();
         }
 
-        FutureModule newFuture = new FutureModule(identifier);
-        futureModule = moduleMap.putIfAbsent(identifier, newFuture);
+        FutureModule newFuture = new FutureModule(name);
+        futureModule = moduleMap.putIfAbsent(name, newFuture);
         if (futureModule != null) {
             return futureModule.getModule();
         }
@@ -337,39 +491,49 @@ public class ModuleLoader {
         boolean ok = false;
         try {
             final ModuleLogger log = Module.log;
-            log.trace("Locally loading module %s from %s", identifier, this);
+            log.trace("Locally loading module %s from %s", name, this);
             final long startTime = Metrics.getCurrentCPUTime();
-            final ModuleSpec moduleSpec = findModule(identifier);
+            final ModuleSpec moduleSpec = findModule(name);
             loadTime.addAndGet(Metrics.getCurrentCPUTime() - startTime);
             if (moduleSpec == null) {
-                log.trace("Module %s not found from %s", identifier, this);
+                log.trace("Module %s not found from %s", name, this);
                 return null;
             }
-            if (! moduleSpec.getModuleIdentifier().equals(identifier)) {
+            if (! moduleSpec.getName().equals(name)) {
                 throw new ModuleLoadException("Module loader found a module with the wrong name");
             }
             final Module module;
-            if ( moduleSpec instanceof AliasModuleSpec) {
-                final ModuleIdentifier aliasTarget = ((AliasModuleSpec) moduleSpec).getAliasTarget();
-                try {
-                    newFuture.setModule(module = loadModuleLocal(aliasTarget));
-                } catch (RuntimeException e) {
-                    log.trace(e, "Failed to load module %s (alias for %s)", identifier, aliasTarget);
-                    throw e;
-                } catch (Error e) {
-                    log.trace(e, "Failed to load module %s (alias for %s)", identifier, aliasTarget);
-                    throw e;
+            if (moduleSpec instanceof AliasModuleSpec) {
+                final AliasModuleSpec aliasModuleSpec = (AliasModuleSpec) moduleSpec;
+                Module aliasedModule;
+                while (true) {
+                    aliasedModule = loadModuleLocal(aliasModuleSpec.getAliasName());
+                    if (aliasedModule == null) {
+                        throw new ModuleLoadException("Alias module " + name + " is referencing not existing module");
+                    }
+                    final Set<String> aliases = aliasedModule.aliases;
+                    if (aliases == null) continue; // race condition detected, retry
+                    synchronized (aliases) {
+                        if (aliases == aliasedModule.aliases) {
+                            // no race-condition with removal, go ahead
+                            aliases.add(moduleSpec.getName());
+                            newFuture.setModule(module = aliasedModule);
+                            log.trace("Added module %s as alias of %s from %s", moduleSpec.getName(), aliasModuleSpec.getAliasName(), this);
+                            ok = true;
+                            break;
+                        }
+                    }
                 }
             } else {
                 module = defineModule((ConcreteModuleSpec) moduleSpec, newFuture);
+                log.trace("Loaded module %s from %s", name, this);
+                ok = true;
             }
-            log.trace("Loaded module %s from %s", identifier, this);
-            ok = true;
             return module;
         } finally {
             if (! ok) {
                 newFuture.setModule(null);
-                moduleMap.remove(identifier, newFuture);
+                moduleMap.remove(name, newFuture);
             }
         }
     }
@@ -381,9 +545,23 @@ public class ModuleLoader {
      *
      * @param identifier the module identifier
      * @return the module, or {@code null} if it wasn't found
+     * @deprecated Use {@link #findLoadedModuleLocal(String)} instead.
      */
+    @Deprecated
     protected final Module findLoadedModuleLocal(ModuleIdentifier identifier) {
-        FutureModule futureModule = moduleMap.get(identifier);
+        return findLoadedModuleLocal(identifier.toString());
+    }
+
+    /**
+     * Find an already-loaded module, returning {@code null} if the module isn't currently loaded.  May block
+     * while the loaded state of the module is in question (if the module is being concurrently loaded from another
+     * thread, for example).
+     *
+     * @param name the module identifier
+     * @return the module, or {@code null} if it wasn't found
+     */
+    protected final Module findLoadedModuleLocal(String name) {
+        FutureModule futureModule = moduleMap.get(name);
         if (futureModule != null) {
             try {
                 return futureModule.getModule();
@@ -407,21 +585,89 @@ public class ModuleLoader {
      * @param module the module to unload
      * @throws SecurityException if an attempt is made to unload a module which does not belong to this module loader
      * @throws SecurityException if the module was not defined by this module loader
+     * @deprecated use {@link #unloadModuleLocal(String,Module)} instead
      */
+    @Deprecated
     protected final void unloadModuleLocal(Module module) throws SecurityException {
+        if (module == null) {
+            throw new NullPointerException("Module parameter cannot be null");
+        }
+        unloadModuleLocal(module.getName(), module);
+    }
+
+    /**
+     * Unload a module from this module loader.  Note that this has no effect on existing modules which refer to the
+     * module being unloaded.  Also, only modules from the current module loader can be unloaded.  Unloading the same
+     * module more than once has no additional effect.  This method only removes the mapping for the module; any running
+     * threads which are currently accessing or linked to the module will continue to function, however attempts to load
+     * this module will fail until a new module is loaded with the same name.  Once this happens, if all references to
+     * the previous module are not cleared, the same module may be loaded more than once, causing possible class duplication
+     * and class cast exceptions if proper care is not taken.
+     *
+     * If the given {@code moduleName} is an alias of the given {@code module}
+     * then only alias mapping to {@code module} is removed. Aliased {@code module}
+     * with its canonical name and other aliases mapping is left intact.
+     *
+     * If the given {@code moduleName} is the canonical {@code module} name
+     * then all its alias plus canonical {@code moduleName} association mappings are removed.
+     *
+     * @param moduleId the module identifier
+     * @param module the module to unload
+     * @return true iff module was unloaded
+     * @throws SecurityException if an attempt is made to unload a module which does not belong to this module loader
+     * @throws SecurityException if the module was not defined by this module loader
+     */
+    protected final boolean unloadModuleLocal(final String moduleId, final Module module) throws SecurityException {
+        if (moduleId == null || module == null) {
+            throw new NullPointerException("Method parameters cannot be null");
+        }
         final ModuleLoader moduleLoader = module.getModuleLoader();
         if (moduleLoader != this) {
             throw new SecurityException("Attempted to unload " + module + " from a different module loader");
         }
-        final ModuleIdentifier id = module.getIdentifier();
-        final FutureModule futureModule = moduleMap.get(id);
+        final FutureModule futureModule = moduleMap.get(moduleId);
         if (futureModule != null && futureModule.module == module) {
-            moduleMap.remove(id, futureModule);
+            final Set<String> aliases = module.aliases;
+            if (aliases != null) {
+                synchronized (aliases) {
+                    if (aliases.remove(moduleId)) {
+                        // unregistered alias from aliased module
+                    } else {
+                        // cleaning aliased module and all its aliases
+                        module.aliases = null;
+                        for (String alias : aliases) {
+                            moduleMap.remove(alias);
+                        }
+                    }
+                }
+            }
+            return moduleMap.remove(moduleId, futureModule);
+        }
+        return false;
+    }
+
+    /**
+     * This compatibility method just calls {@code findModule(methodIdentifier.toString())} and is usually not used;
+     * if however a legacy subclass overrides this method, then it will be called instead of {@link #findModule(String)},
+     * and that method will redirect to this one.
+     *
+     * @deprecated Use {@link #findModule(String)} instead.
+     *
+     * @param moduleIdentifier the module identifier
+     * @return the module specification
+     * @throws ModuleLoadException
+     */
+    @Deprecated
+    protected ModuleSpec findModule(final ModuleIdentifier moduleIdentifier) throws ModuleLoadException {
+        if (overridesFindModuleByIdentifier) {
+            return findModule0(moduleIdentifier.toString());
+        } else {
+            return findModule(moduleIdentifier.toString());
         }
     }
 
     /**
-     * Find a Module's specification in this ModuleLoader by its identifier.  This can be overriden by sub-classes to
+     * Find a Module's specification in this ModuleLoader by its name.  This can be overriden by sub-classes to
      * implement the Module loading strategy for this loader.  The default implementation iterates the module finders
      * provided during construction.
      * <p/>
@@ -429,14 +675,22 @@ public class ModuleLoader {
      * null}. If the module is found but some problem occurred (for example, a transitive dependency failed to load)
      * then this method should throw a {@link ModuleLoadException} of the relevant type.
      *
-     * @param moduleIdentifier the module identifier
+     * @param name the module name
      * @return the module specification, or {@code null} if no module is found with the given identifier
      * @throws ModuleLoadException if any problems occur finding the module
      */
-    protected ModuleSpec findModule(final ModuleIdentifier moduleIdentifier) throws ModuleLoadException {
+    protected ModuleSpec findModule(final String name) throws ModuleLoadException {
+        if (overridesFindModuleByIdentifier) {
+            return findModule(ModuleIdentifier.fromString(name));
+        } else {
+            return findModule0(name);
+        }
+    }
+
+    private ModuleSpec findModule0(final String name) throws ModuleLoadException {
         for (ModuleFinder finder : finders) {
             if (finder != null) {
-                final ModuleSpec spec = finder.findModule(moduleIdentifier, this);
+                final ModuleSpec spec = finder.findModule(name, this);
                 if (spec != null) {
                     return spec;
                 }
@@ -455,7 +709,7 @@ public class ModuleLoader {
     }
 
     /**
-     * Defines a Module based on a specification.  May only be called from {@link #loadModuleLocal(ModuleIdentifier)}.
+     * Defines a Module based on a specification.  May only be called from {@link #loadModuleLocal(String)}.
      *
      * @param moduleSpec The module specification to create the Module from
      * @param futureModule the future module to populate
@@ -467,20 +721,20 @@ public class ModuleLoader {
             return doPrivileged(new PrivilegedExceptionAction<Module>() {
                 public Module run() throws Exception {
                     final ModuleLogger log = Module.log;
-                    final ModuleIdentifier moduleIdentifier = moduleSpec.getModuleIdentifier();
+                    final String name = moduleSpec.getName();
 
                     final Module module = new Module(moduleSpec, ModuleLoader.this);
                     module.getClassLoaderPrivate().recalculate();
                     module.setDependencies(moduleSpec.getDependenciesInternal());
-                    log.moduleDefined(moduleIdentifier, ModuleLoader.this);
+                    log.moduleDefined(name, ModuleLoader.this);
                     try {
                         futureModule.setModule(module);
                         return module;
                     } catch (RuntimeException e) {
-                        log.trace(e, "Failed to load module %s", moduleIdentifier);
+                        log.trace(e, "Failed to load module %s", name);
                         throw e;
                     } catch (Error e) {
-                        log.trace(e, "Failed to load module %s", moduleIdentifier);
+                        log.trace(e, "Failed to load module %s", name);
                         throw e;
                     }
                 }
@@ -625,14 +879,15 @@ public class ModuleLoader {
         if (Metrics.ENABLED) classCount.getAndIncrement();
     }
 
-    private static final class FutureModule {
-        private static final Object NOT_FOUND = new Object();
+    static final Object NOT_FOUND = new Object();
 
-        final ModuleIdentifier identifier;
+    final class FutureModule {
+
+        final String name;
         volatile Object module;
 
-        FutureModule(final ModuleIdentifier identifier) {
-            this.identifier = identifier;
+        FutureModule(final String name) {
+            this.name = name;
         }
 
         Module getModule() throws ModuleNotFoundException {
@@ -648,7 +903,7 @@ public class ModuleLoader {
                         }
                     }
                 }
-                if (module == NOT_FOUND) throw new ModuleNotFoundException(identifier.toString());
+                if (module == NOT_FOUND) return null;
                 return (Module) module;
             } finally {
                 if (intr) Thread.currentThread().interrupt();
@@ -709,13 +964,10 @@ public class ModuleLoader {
         }
 
         public List<String> queryLoadedModuleNames() {
-            ModuleLoader loader = getModuleLoader();
-            final Set<ModuleIdentifier> identifiers = loader.moduleMap.keySet();
-            final ArrayList<String> list = new ArrayList<String>(identifiers.size());
-            for (ModuleIdentifier identifier : identifiers) {
-                list.add(identifier.toString());
-            }
-            Collections.sort(list);
+            final ModuleLoader loader = getModuleLoader();
+            final Set<String> names = loader.moduleMap.keySet();
+            final ArrayList<String> list = new ArrayList<>(names);
+            list.sort(Comparator.naturalOrder());
             return list;
         }
 
@@ -777,11 +1029,11 @@ public class ModuleLoader {
                 sm.checkPermission(MODULE_UNLOAD_ANY_PERM);
             }
             final ModuleLoader loader = getModuleLoader();
-            final Module module = loader.findLoadedModuleLocal(ModuleIdentifier.fromString(name));
+            final Module module = loader.findLoadedModuleLocal(name);
             if (module == null) {
                 return false;
             } else {
-                loader.unloadModuleLocal(module);
+                loader.unloadModuleLocal(name, module);
                 return true;
             }
         }
@@ -834,7 +1086,7 @@ public class ModuleLoader {
                     info = new DependencyInfo(dependencyType, exportFilter, importFilter, null, null, false, localDependency.getLocalLoader().toString(), pathList);
                 } else if (dependency instanceof ModuleDependency) {
                     final ModuleDependency moduleDependency = (ModuleDependency) dependency;
-                    info = new DependencyInfo(dependencyType, exportFilter, importFilter, moduleDependency.getModuleLoader().mxBean, moduleDependency.getIdentifier().toString(), moduleDependency.isOptional(), null, null);
+                    info = new DependencyInfo(dependencyType, exportFilter, importFilter, moduleDependency.getModuleLoader().mxBean, moduleDependency.getName(), moduleDependency.isOptional(), null, null);
                 } else {
                     info = new DependencyInfo(dependencyType, exportFilter, importFilter, null, null, false, null, null);
                 }
@@ -854,7 +1106,7 @@ public class ModuleLoader {
             final ResourceLoader[] loaders = classLoader.getResourceLoaders();
             final ArrayList<ResourceLoaderInfo> list = new ArrayList<ResourceLoaderInfo>(loaders.length);
             for (ResourceLoader resourceLoader : loaders) {
-                list.add(new ResourceLoaderInfo(resourceLoader.getClass().getName(), resourceLoader.getLocation().toString(), new ArrayList<>(resourceLoader.getPaths())));
+                list.add(new ResourceLoaderInfo(resourceLoader.getClass().getName(), String.valueOf(resourceLoader.getLocation()), new ArrayList<>(resourceLoader.getPaths())));
             }
             return list;
         }
@@ -899,7 +1151,7 @@ public class ModuleLoader {
 
         private Module loadModule(final String name, final ModuleLoader loader) {
             try {
-                final Module module = loader.findLoadedModuleLocal(ModuleIdentifier.fromString(name));
+                final Module module = loader.findLoadedModuleLocal(name);
                 if (module == null) {
                     throw new IllegalArgumentException("Module " + name + " not found");
                 }
