@@ -34,6 +34,7 @@ import java.util.Arrays;
 import java.util.Collections;
 import java.util.Enumeration;
 import java.util.HashSet;
+import java.util.Iterator;
 import java.util.List;
 import java.util.Locale;
 import java.util.Set;
@@ -43,7 +44,11 @@ import java.util.zip.ZipFile;
 
 import org.jboss.modules.AssertionSetting;
 import org.jboss.modules.DependencySpec;
+import org.jboss.modules.LocalDependencySpecBuilder;
+import org.jboss.modules.ModuleDependencySpec;
+import org.jboss.modules.ModuleDependencySpecBuilder;
 import org.jboss.modules.Version;
+import org.jboss.modules.VersionDetection;
 import org.jboss.modules.maven.ArtifactCoordinates;
 import org.jboss.modules.maven.MavenArtifactUtil;
 import org.jboss.modules.ModuleIdentifier;
@@ -84,12 +89,15 @@ import static org.jboss.modules.xml.XmlPullParser.TEXT;
  */
 public final class ModuleXmlParser {
 
+    static final ModuleDependencySpec DEP_JAVA_SE = new ModuleDependencySpecBuilder().setName("java.se").build();
+    static final ModuleDependencySpec DEP_JDK_UNSUPPORTED = new ModuleDependencySpecBuilder().setName("jdk.unsupported").setOptional(true).build();
+
     /**
      * A factory for resource roots, based on a root path, loader path, and loader name.  Normally it is sufficient to
      * accept the default.
      */
     public interface ResourceRootFactory {
-        ResourceLoader createResourceLoader(final String rootPath, final String loaderPath, final String loaderName) throws IOException;
+        ResourceLoader createResourceLoader(final String rootPath, final String loaderPath, @Deprecated final String loaderName) throws IOException;
 
         /**
          * Get the default resource root factory.
@@ -111,6 +119,8 @@ public final class ModuleXmlParser {
     // there is no 1.4
     private static final String MODULE_1_5 = "urn:jboss:module:1.5";
     private static final String MODULE_1_6 = "urn:jboss:module:1.6";
+    private static final String MODULE_1_7 = "urn:jboss:module:1.7";
+    private static final String MODULE_1_8 = "urn:jboss:module:1.8";
 
     private static final String E_MODULE = "module";
     private static final String E_ARTIFACT = "artifact";
@@ -135,6 +145,12 @@ public final class ModuleXmlParser {
     private static final String E_PROPERTY = "property";
     private static final String E_PERMISSIONS = "permissions";
     private static final String E_GRANT = "grant";
+    private static final String E_CONDITIONS = "conditions";
+    private static final String E_PROPERTY_EQUAL = "property-equal";
+    private static final String E_PROPERTY_NOT_EQUAL = "property-not-equal";
+    private static final String E_PROVIDES = "provides";
+    private static final String E_SERVICE = "service";
+    private static final String E_WITH_CLASS = "with-class";
 
     private static final String A_NAME = "name";
     private static final String A_SLOT = "slot";
@@ -343,7 +359,7 @@ public final class ModuleXmlParser {
         return new XmlPullParserException(b.toString(), reader, null);
     }
 
-    private static XmlPullParserException unknownAttribute(final XmlPullParser parser, final int index) {
+    static XmlPullParserException unknownAttribute(final XmlPullParser parser, final int index) {
         final String namespace = parser.getAttributeNamespace(index);
         final String prefix = parser.getAttributePrefix(index);
         final String name = parser.getAttributeName(index);
@@ -377,13 +393,23 @@ public final class ModuleXmlParser {
             case MODULE_1_3:
             case MODULE_1_5:
             case MODULE_1_6:
+            case MODULE_1_7:
+            case MODULE_1_8:
                 break;
             default: throw unexpectedContent(reader);
         }
     }
 
     private static boolean atLeast1_6(final XmlPullParser reader) {
-            return MODULE_1_6.equals(reader.getNamespace());
+        return MODULE_1_6.equals(reader.getNamespace()) || atLeast1_7(reader);
+    }
+
+    private static boolean atLeast1_7(final XmlPullParser reader) {
+        return MODULE_1_7.equals(reader.getNamespace()) || atLeast1_8(reader);
+    }
+
+    private static boolean atLeast1_8(final XmlPullParser reader) {
+        return MODULE_1_8.equals(reader.getNamespace());
     }
 
     private static void assertNoAttributes(final XmlPullParser reader) throws XmlPullParserException {
@@ -412,8 +438,7 @@ public final class ModuleXmlParser {
                     final String element = reader.getName();
                     switch (element) {
                         case E_MODULE: {
-                            final ModuleSpec.Builder specBuilder = ModuleSpec.build(moduleName);
-                            parseModuleContents(mavenResolver, reader, factory, moduleLoader, moduleName, specBuilder, rootPath);
+                            final ModuleSpec.Builder specBuilder = parseModuleContents(mavenResolver, reader, factory, moduleLoader, moduleName, rootPath);
                             parseEndDocument(reader);
                             return specBuilder.create();
                         }
@@ -448,8 +473,7 @@ public final class ModuleXmlParser {
                     final String element = reader.getName();
                     switch (element) {
                         case E_MODULE: {
-                            final ModuleSpec.Builder specBuilder = ModuleSpec.build(moduleName);
-                            parseModuleContents(mavenResolver, reader, factory, moduleLoader, moduleName, specBuilder, rootPath);
+                            final ModuleSpec.Builder specBuilder = parseModuleContents(mavenResolver, reader, factory, moduleLoader, moduleName, rootPath);
                             parseEndDocument(reader);
                             return specBuilder.create();
                         }
@@ -498,13 +522,15 @@ public final class ModuleXmlParser {
         if (! required.isEmpty()) {
             throw missingAttributes(reader, required);
         }
-        if (noSlots) {
-            if (! moduleName.equals(name)) {
-                throw invalidModuleName(reader, moduleName);
-            }
-        } else {
-            if (! ModuleIdentifier.fromString(moduleName).equals(ModuleIdentifier.create(name, slot))) {
-                throw invalidModuleName(reader, moduleName);
+        if (moduleName != null) {
+            if (noSlots) {
+                if (! moduleName.equals(name)) {
+                    throw invalidModuleName(reader, moduleName);
+                }
+            } else {
+                if (! ModuleIdentifier.fromString(moduleName).equals(ModuleIdentifier.create(name, slot))) {
+                    throw invalidModuleName(reader, moduleName);
+                }
             }
         }
         int eventType;
@@ -513,9 +539,9 @@ public final class ModuleXmlParser {
             switch (eventType) {
                 case END_TAG: {
                     if (noSlots) {
-                        return ModuleSpec.buildAlias(moduleName, targetName).create();
+                        return ModuleSpec.buildAlias(name, targetName).create();
                     } else {
-                        return ModuleSpec.buildAlias(ModuleIdentifier.fromString(moduleName), ModuleIdentifier.create(targetName, targetSlot)).create();
+                        return ModuleSpec.buildAlias(ModuleIdentifier.create(name, slot), ModuleIdentifier.create(targetName, targetSlot)).create();
                     }
                 }
                 default: {
@@ -577,20 +603,20 @@ public final class ModuleXmlParser {
         return permissions;
     }
 
-    private static void parseModuleContents(final MavenResolver mavenResolver, final XmlPullParser reader, final ResourceRootFactory factory, final ModuleLoader moduleLoader, final String moduleName, final ModuleSpec.Builder specBuilder, final String rootPath) throws XmlPullParserException, IOException {
+    private static ModuleSpec.Builder parseModuleContents(final MavenResolver mavenResolver, final XmlPullParser reader, final ResourceRootFactory factory, final ModuleLoader moduleLoader, final String moduleName, final String rootPath) throws XmlPullParserException, IOException {
         final int count = reader.getAttributeCount();
         String name = null;
         String slot = null;
-        boolean noSlots = atLeast1_6(reader);
+        boolean is1_6 = atLeast1_6(reader);
         Version version = null;
-        final Set<String> required = noSlots ? new HashSet<>(LIST_A_NAME) : new HashSet<>(LIST_A_NAME);
+        final Set<String> required = is1_6 ? new HashSet<>(LIST_A_NAME) : new HashSet<>(LIST_A_NAME);
         for (int i = 0; i < count; i ++) {
             validateAttributeNamespace(reader, i);
             final String attribute = reader.getAttributeName(i);
             required.remove(attribute);
             switch (attribute) {
                 case A_NAME:    name = reader.getAttributeValue(i); break;
-                case A_SLOT:    if (noSlots) throw unknownAttribute(reader, i); else slot = reader.getAttributeValue(i); break;
+                case A_SLOT:    if (is1_6) throw unknownAttribute(reader, i); else slot = reader.getAttributeValue(i); break;
                 case A_VERSION:
                     try {
                         version = Version.parse(reader.getAttributeValue(i));
@@ -604,19 +630,26 @@ public final class ModuleXmlParser {
         if (! required.isEmpty()) {
             throw missingAttributes(reader, required);
         }
-        if (noSlots) {
-            if (! specBuilder.getName().equals(name)) {
-                throw invalidModuleName(reader, specBuilder.getName());
-            }
+        final String realModuleName;
+        if (is1_6) {
+            realModuleName = name;
         } else {
-            if (! specBuilder.getIdentifier().equals(ModuleIdentifier.create(name, slot))) {
-                throw invalidModuleName(reader, specBuilder.getIdentifier().toString());
-            }
+            realModuleName = ModuleIdentifier.create(name, slot).toString();
         }
+        if (moduleName != null && ! realModuleName.equals(moduleName)) {
+            throw invalidModuleName(reader, realModuleName);
+        }
+        final ModuleSpec.Builder specBuilder = ModuleSpec.build(realModuleName);
         specBuilder.setVersion(version);
         // xsd:all
         MultiplePathFilterBuilder exportsBuilder = PathFilters.multiplePathFilterBuilder(true);
         ArrayList<DependencySpec> dependencies = new ArrayList<>();
+        final boolean is1_8 = atLeast1_8(reader);
+        if (! is1_8) {
+            // add default system dependencies
+            specBuilder.addDependency(DEP_JAVA_SE);
+            specBuilder.addDependency(DEP_JDK_UNSUPPORTED);
+        }
         Set<String> visited = new HashSet<>();
         int eventType;
         boolean gotPerms = false;
@@ -624,12 +657,14 @@ public final class ModuleXmlParser {
             eventType = reader.nextTag();
             switch (eventType) {
                 case END_TAG: {
-                    specBuilder.addDependency(DependencySpec.createLocalDependencySpec(PathFilters.acceptAll(), exportsBuilder.create()));
+                    specBuilder.addDependency(new LocalDependencySpecBuilder()
+                        .setExportFilter(exportsBuilder.create())
+                        .build());
                     for (DependencySpec dependency : dependencies) {
                         specBuilder.addDependency(dependency);
                     }
                     if (! gotPerms) specBuilder.setPermissionCollection(DEFAULT_PERMISSION_COLLECTION);
-                    return;
+                    return specBuilder;
                 }
                 case START_TAG: {
                     validateNamespace(reader);
@@ -644,7 +679,8 @@ public final class ModuleXmlParser {
                         case E_MAIN_CLASS:   parseMainClass(reader, specBuilder); break;
                         case E_RESOURCES:    parseResources(mavenResolver, factory, rootPath, reader, specBuilder); break;
                         case E_PROPERTIES:   parseProperties(reader, specBuilder); break;
-                        case E_PERMISSIONS:  parsePermissions(reader, moduleLoader, moduleName, specBuilder); gotPerms = true; break;
+                        case E_PERMISSIONS:  parsePermissions(reader, moduleLoader, realModuleName, specBuilder); gotPerms = true; break;
+                        case E_PROVIDES:     if (is1_8) parseProvidesType(reader, specBuilder); else throw unexpectedContent(reader); break;
                         default: throw unexpectedContent(reader);
                     }
                     break;
@@ -670,7 +706,7 @@ public final class ModuleXmlParser {
                     validateNamespace(reader);
                     switch (reader.getName()) {
                         case E_MODULE: parseModuleDependency(reader, dependencies); break;
-                        case E_SYSTEM: parseSystemDependency(reader, dependencies); break;
+                        case E_SYSTEM: if (! atLeast1_8(reader)) { parseSystemDependency(reader, dependencies); break; }
                         default: throw unexpectedContent(reader);
                     }
                     break;
@@ -746,11 +782,12 @@ public final class ModuleXmlParser {
                         importBuilder.addFilter(PathFilters.getMetaInfFilter(), false);
                         importFilter = importBuilder.create();
                     }
-                    if (noSlots) {
-                        dependencies.add(DependencySpec.createModuleDependencySpec(importFilter, exportFilter, null, name, optional));
-                    } else {
-                        dependencies.add(DependencySpec.createModuleDependencySpec(importFilter, exportFilter, null, ModuleIdentifier.create(name, slot), optional));
-                    }
+                    dependencies.add(new ModuleDependencySpecBuilder()
+                        .setImportFilter(importFilter)
+                        .setExportFilter(exportFilter)
+                        .setName(noSlots ? name : ModuleIdentifier.create(name, slot).toString())
+                        .setOptional(optional)
+                        .build());
                     return;
                 }
                 case START_TAG: {
@@ -811,7 +848,7 @@ public final class ModuleXmlParser {
         }
     }
 
-    private static void parseMainClass(final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
+    private static String parseClassNameType(final XmlPullParser reader) throws IOException, XmlPullParserException {
         String name = null;
         final Set<String> required = new HashSet<>(LIST_A_NAME);
         final int count = reader.getAttributeCount();
@@ -827,19 +864,38 @@ public final class ModuleXmlParser {
         if (! required.isEmpty()) {
             throw missingAttributes(reader, required);
         }
-        specBuilder.setMainClass(name);
         // consume remainder of element
         parseNoContent(reader);
+        return name;
+    }
+
+    private static void parseMainClass(final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
+        specBuilder.setMainClass(parseClassNameType(reader));
     }
 
     private static void parseResources(final MavenResolver mavenResolver, final ResourceRootFactory factory, final String rootPath, final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
         assertNoAttributes(reader);
+        final List<Version> detectedVersions = new ArrayList<>();
         // xsd:choice
         int eventType;
         for (;;) {
             eventType = reader.nextTag();
             switch (eventType) {
                 case END_TAG: {
+                    final Version specifiedVersion = specBuilder.getVersion();
+                    if (specifiedVersion == null) {
+                        final Iterator<Version> iterator = detectedVersions.iterator();
+                        if (iterator.hasNext()) {
+                            Version guess = iterator.next();
+                            while (iterator.hasNext()) {
+                                if (! guess.equals(iterator.next())) {
+                                    guess = null;
+                                    break;
+                                }
+                            }
+                            if (guess != null) specBuilder.setVersion(guess);
+                        }
+                    }
                     specBuilder.addResourceRoot(ResourceLoaderSpec.createResourceLoaderSpec(new NativeLibraryResourceLoader(new File(rootPath, "lib")), PathFilters.rejectAll()));
                     return;
                 }
@@ -847,11 +903,13 @@ public final class ModuleXmlParser {
                     validateNamespace(reader);
                     switch (reader.getName()) {
                         case E_RESOURCE_ROOT: {
-                            parseResourceRoot(factory, rootPath, reader, specBuilder);
+                            final Version version = parseResourceRoot(factory, rootPath, reader, specBuilder);
+                            if (version != null) detectedVersions.add(version);
                             break;
                         }
                         case E_ARTIFACT: {
-                            parseArtifact(mavenResolver, reader, specBuilder);
+                            final Version version = parseArtifact(mavenResolver, reader, specBuilder);
+                            if (version != null) detectedVersions.add(version);
                             break;
                         }
                         case E_NATIVE_ARTIFACT: {
@@ -897,20 +955,31 @@ public final class ModuleXmlParser {
             throw missingAttributes(reader, required);
         }
 
+        final SystemPropertyConditionBuilder conditionBuilder = new SystemPropertyConditionBuilder();
+        final Set<String> encountered = new HashSet<>();
         int eventType;
         for (;;) {
             eventType = reader.nextTag();
             switch (eventType) {
                 case END_TAG: {
                     try {
-                        createMavenNativeArtifactLoader(mavenResolver, name, reader, specBuilder);
+                        if (conditionBuilder.resolve()) {
+                            createMavenNativeArtifactLoader(mavenResolver, name, reader, specBuilder);
+                        }
                     } catch (IOException e) {
                         throw new XmlPullParserException(String.format("Failed to add artifact '%s'", name), reader, e);
                     }
                     return;
                 }
                 case START_TAG: {
-                    throw unexpectedContent(reader);
+                    validateNamespace(reader);
+                    final String element = reader.getName();
+                    if (! encountered.add(element)) throw unexpectedContent(reader);
+                    switch (element) {
+                        case E_CONDITIONS: parseConditions(reader, conditionBuilder); break;
+                        default: throw unexpectedContent(reader);
+                    }
+                    break;
                 }
                 default: {
                     throw unexpectedContent(reader);
@@ -919,7 +988,7 @@ public final class ModuleXmlParser {
         }
     }
 
-    private static void parseArtifact(final MavenResolver mavenResolver, final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
+    private static Version parseArtifact(final MavenResolver mavenResolver, final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
         String name = null;
         final Set<String> required = new HashSet<>(LIST_A_NAME);
         final int count = reader.getAttributeCount();
@@ -937,6 +1006,7 @@ public final class ModuleXmlParser {
         }
 
         final MultiplePathFilterBuilder filterBuilder = PathFilters.multiplePathFilterBuilder(true);
+        final SystemPropertyConditionBuilder conditionBuilder = new SystemPropertyConditionBuilder();
         final ResourceLoader resourceLoader;
 
         final Set<String> encountered = new HashSet<>();
@@ -945,14 +1015,24 @@ public final class ModuleXmlParser {
             eventType = reader.nextTag();
             switch (eventType) {
                 case END_TAG: {
-                    try {
-                        resourceLoader = MavenArtifactUtil.createMavenArtifactLoader(mavenResolver, name);
-                    } catch (IOException e) {
-                        throw new XmlPullParserException(String.format("Failed to add artifact '%s'", name), reader, e);
+                    if(conditionBuilder.resolve()) {
+                        final ArtifactCoordinates coordinates;
+                        try {
+                            coordinates = ArtifactCoordinates.fromString(name);
+                            resourceLoader = MavenArtifactUtil.createMavenArtifactLoader(mavenResolver, coordinates, name);
+                        } catch (IOException | IllegalArgumentException e) {
+                            throw new XmlPullParserException(String.format("Failed to add artifact '%s'", name), reader, e);
+                        }
+                        if (resourceLoader == null) throw new XmlPullParserException(String.format("Failed to resolve artifact '%s'", name), reader, null);
+                        specBuilder.addResourceRoot(ResourceLoaderSpec.createResourceLoaderSpec(resourceLoader, filterBuilder.create()));
+                        final String version = coordinates.getVersion();
+                        try {
+                            return Version.parse(version);
+                        } catch (IllegalArgumentException ignored) {
+                            return null;
+                        }
                     }
-                    if (resourceLoader == null) throw new XmlPullParserException(String.format("Failed to resolve artifact '%s'", name), reader, null);
-                    specBuilder.addResourceRoot(ResourceLoaderSpec.createResourceLoaderSpec(resourceLoader, filterBuilder.create()));
-                    return;
+                    return null;
                 }
                 case START_TAG: {
                     validateNamespace(reader);
@@ -960,6 +1040,7 @@ public final class ModuleXmlParser {
                     if (! encountered.add(element)) throw unexpectedContent(reader);
                     switch (element) {
                         case E_FILTER: parseFilterList(reader, filterBuilder); break;
+                        case E_CONDITIONS: parseConditions(reader, conditionBuilder); break;
                         default: throw unexpectedContent(reader);
                     }
                     break;
@@ -971,17 +1052,18 @@ public final class ModuleXmlParser {
         }
     }
 
-    private static void parseResourceRoot(final ResourceRootFactory factory, final String rootPath, final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
+    private static Version parseResourceRoot(final ResourceRootFactory factory, final String rootPath, final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
         String name = null;
         String path = null;
         final Set<String> required = new HashSet<>(LIST_A_PATH);
         final int count = reader.getAttributeCount();
+        final boolean is1_8 = atLeast1_8(reader);
         for (int i = 0; i < count; i ++) {
             validateAttributeNamespace(reader, i);
             final String attribute = reader.getAttributeName(i);
             required.remove(attribute);
             switch (attribute) {
-                case A_NAME: name = reader.getAttributeValue(i); break;
+                case A_NAME: if (!is1_8) name = reader.getAttributeValue(i); else throw unknownAttribute(reader, i); break;
                 case A_PATH: path = reader.getAttributeValue(i); break;
                 default: throw unknownAttribute(reader, i);
             }
@@ -989,9 +1071,11 @@ public final class ModuleXmlParser {
         if (! required.isEmpty()) {
             throw missingAttributes(reader, required);
         }
-        if (name == null) name = path;
+        if (name == null) name = is1_8 ? "unnamed" : path;
 
         final MultiplePathFilterBuilder filterBuilder = PathFilters.multiplePathFilterBuilder(true);
+        final SystemPropertyConditionBuilder conditionBuilder = new SystemPropertyConditionBuilder();
+
         final ResourceLoader resourceLoader;
 
         final Set<String> encountered = new HashSet<>();
@@ -1000,13 +1084,21 @@ public final class ModuleXmlParser {
             eventType = reader.nextTag();
             switch (eventType) {
                 case END_TAG: {
-                    try {
-                        resourceLoader = factory.createResourceLoader(rootPath, path, name);
-                    } catch (IOException e) {
-                        throw new XmlPullParserException(String.format("Failed to add resource root '%s' at path '%s'", name, path), reader, e);
+                    if(conditionBuilder.resolve()) {
+                        try {
+                            resourceLoader = factory.createResourceLoader(rootPath, path, name);
+                        } catch (IOException e) {
+                            throw new XmlPullParserException(String.format("Failed to add resource root '%s' at path '%s'", name, path), reader, e);
+                        }
+                        specBuilder.addResourceRoot(ResourceLoaderSpec.createResourceLoaderSpec(resourceLoader, filterBuilder.create()));
+                        if (specBuilder.getVersion() == null) {
+                            return VersionDetection.detectVersion(resourceLoader);
+                        } else {
+                            return null;
+                        }
+                    } else {
+                        return null;
                     }
-                    specBuilder.addResourceRoot(ResourceLoaderSpec.createResourceLoaderSpec(resourceLoader, filterBuilder.create()));
-                    return;
                 }
                 case START_TAG: {
                     validateNamespace(reader);
@@ -1014,6 +1106,7 @@ public final class ModuleXmlParser {
                     if (! encountered.add(element)) throw unexpectedContent(reader);
                     switch (element) {
                         case E_FILTER: parseFilterList(reader, filterBuilder); break;
+                        case E_CONDITIONS: parseConditions(reader, conditionBuilder); break;
                         default: throw unexpectedContent(reader);
                     }
                     break;
@@ -1025,6 +1118,58 @@ public final class ModuleXmlParser {
         }
     }
 
+    private static void parseConditions(XmlPullParser reader, SystemPropertyConditionBuilder conditionBuilder) throws XmlPullParserException, IOException {
+        if(!atLeast1_7(reader)) {
+            throw unexpectedContent(reader);
+        }
+        assertNoAttributes(reader);
+        // xsd:choice
+        int eventType;
+        for (;;) {
+            eventType = reader.nextTag();
+            switch (eventType) {
+                case END_TAG: {
+                    return;
+                }
+                case START_TAG: {
+                    validateNamespace(reader);
+                    switch (reader.getName()) {
+                        case E_PROPERTY_EQUAL: parseConditionalProperty(reader, true, conditionBuilder); break;
+                        case E_PROPERTY_NOT_EQUAL: parseConditionalProperty(reader, false, conditionBuilder); break;
+                        default: throw unexpectedContent(reader);
+                    }
+                    break;
+                }
+                default: {
+                    throw unexpectedContent(reader);
+                }
+            }
+        }
+    }
+
+    private static void parseConditionalProperty(XmlPullParser reader, boolean equal, SystemPropertyConditionBuilder builder) throws XmlPullParserException, IOException {
+        String name = null;
+        String value = null;
+        final Set<String> required = new HashSet<>(Arrays.asList(A_NAME, A_VALUE));
+        final int count = reader.getAttributeCount();
+        for (int i = 0; i < count; i ++) {
+            validateAttributeNamespace(reader, i);
+            final String attribute = reader.getAttributeName(i);
+            required.remove(attribute);
+            switch (attribute) {
+                case A_NAME: name = reader.getAttributeValue(i); break;
+                case A_VALUE: value = reader.getAttributeValue(i); break;
+                default: throw unknownAttribute(reader, i);
+            }
+        }
+        if (! required.isEmpty()) {
+            throw missingAttributes(reader, required);
+        }
+        builder.add(name, value, equal);
+        // consume remainder of element
+        parseNoContent(reader);
+    }
+
     private static void parseFilterList(final XmlPullParser reader, final MultiplePathFilterBuilder builder) throws XmlPullParserException, IOException {
         assertNoAttributes(reader);
         // xsd:choice
@@ -1244,6 +1389,72 @@ public final class ModuleXmlParser {
         parseNoContent(reader);
     }
 
+    private static void parseProvidesType(final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
+        assertNoAttributes(reader);
+        int eventType;
+        for (;;) {
+            eventType = reader.nextTag();
+            switch (eventType) {
+                case END_TAG: {
+                    return;
+                }
+                case START_TAG: {
+                    validateNamespace(reader);
+                    switch (reader.getName()) {
+                        case E_SERVICE: {
+                            parseProvidedServiceType(reader, specBuilder);
+                            break;
+                        }
+                        default: throw unexpectedContent(reader);
+                    }
+                    break;
+                }
+                default: {
+                    throw unexpectedContent(reader);
+                }
+            }
+        }
+    }
+
+    private static void parseProvidedServiceType(final XmlPullParser reader, final ModuleSpec.Builder specBuilder) throws XmlPullParserException, IOException {
+        String name = null;
+        final int count = reader.getAttributeCount();
+        for (int i = 0; i < count; i ++) {
+            validateAttributeNamespace(reader, i);
+            final String attribute = reader.getAttributeName(i);
+            switch (attribute) {
+                case A_NAME: name = reader.getAttributeValue(i); break;
+                default: throw unknownAttribute(reader, i);
+            }
+        }
+        if (name == null) {
+            throw missingAttributes(reader, Collections.singleton("name"));
+        }
+        int eventType;
+        for (;;) {
+            eventType = reader.nextTag();
+            switch (eventType) {
+                case END_TAG: {
+                    return;
+                }
+                case START_TAG: {
+                    validateNamespace(reader);
+                    switch (reader.getName()) {
+                        case E_WITH_CLASS: {
+                            specBuilder.addProvide(name, parseClassNameType(reader));
+                            break;
+                        }
+                        default: throw unexpectedContent(reader);
+                    }
+                    break;
+                }
+                default: {
+                    throw unexpectedContent(reader);
+                }
+            }
+        }
+    }
+
     private static void expandName(final ModuleLoader moduleLoader, final String moduleName,
             final ArrayList<PermissionFactory> list, String permission, String name, String actions) {
         String expandedName = PolicyExpander.expand(name);
@@ -1375,7 +1586,7 @@ public final class ModuleXmlParser {
                 file = new File(rootPathName, loaderFileName);
             }
             if (file.isDirectory()) {
-                return ResourceLoaders.createFileResourceLoader(loaderName, file);
+                return ResourceLoaders.createPathResourceLoader(loaderName, file.toPath());
             } else {
                 final JarFile jarFile = JDKSpecific.getJarFile(file, true);
                 return ResourceLoaders.createJarResourceLoader(loaderName, jarFile);
